package no.tornado.query;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class ResultMap {
    public enum MapType {PRIMARY_KEY, TABLE_COLUMN, JOIN_COLUMN}

    private Class resultClass;
    private Map<String, ResultMapping> mappings = new LinkedHashMap<>();
    public static final Boolean DEFAULT_STRICT = true;
    public static final Boolean DEFAULT_AUTOMAP = false;
    private Boolean strict = DEFAULT_STRICT;
    private Boolean automap = DEFAULT_AUTOMAP;
    private String tablename;

    private ResultMap() {

    }

    public String describe() {
        StringBuilder s = new StringBuilder();
        s.append("ResultMap " + resultClass.getName() + " for table " + tablename + ":\n\n");
        for (ResultMapping mapping : mappings.values())
            s.append("\t" + mapping.toString() + "\n");

        List<JoinTable> joinTables = uniqueJoinTables();
        if (!joinTables.isEmpty()) {
            s.append("\n");
            for (JoinTable joinTable : joinTables)
                s.append("\t" + joinTable + " ON " + joinTable.getOn() + "\n");
        }

        s.append("\n").append(select());
        s.append("\n").append(insert());
        s.append("\n").append(update());
        return s.toString();
    }

    public Map<String, ResultMapping> getMappings() {
        return mappings;
    }

    public static String skeleton(Class domainClass, String tablename, Connection connection) throws SQLException, IllegalAccessException, InstantiationException {
        StringBuilder s = new StringBuilder("ResultMap " + tablename + "ResultMap = new ResultMap(" + domainClass.getSimpleName() + ".class) {{\n");

        ResultSet rs = connection.createStatement().executeQuery("SELECT * FROM " + tablename);
        ResultSetMetaData m = rs.getMetaData();
        DatabaseMetaData dbm = connection.getMetaData();
        
        ResultSet primaryKeys = dbm.getPrimaryKeys(null, null, tablename);
        List<String> pks = new ArrayList<>();
        while (primaryKeys.next())
            pks.add(primaryKeys.getString("COLUMN_NAME"));

        ResultSet importedKeys = dbm.getImportedKeys(null, null, tablename);
        Map<String, JoinTable> joinTables = new HashMap<>();
        while (importedKeys.next()) {
            String fkColumnName = importedKeys.getString("FKCOLUMN_NAME");
            String pkTableName = importedKeys.getString("PKTABLE_NAME");
            String pkColumnName = importedKeys.getString("PKCOLUMN_NAME");
            JoinTable joinTable = new JoinTable(pkTableName, fkColumnName).on(tablename + "." + fkColumnName + " = " + fkColumnName + "." + pkColumnName);
            joinTable.pkColumn = pkColumnName;
            joinTables.put(fkColumnName, joinTable);
            s.append("\tJoinTable " + fkColumnName + " = new JoinTable(")
            .append("\"" + joinTable.getName() + "\", \"" + joinTable.getAlias() + "\").on(\"" + joinTable.getOn() + "\");\n");
        }
        if (!importedKeys.wasNull())
            s.append("\n");

        Object prototype = domainClass.newInstance();
        s.append("\ttablename(\"").append(tablename).append("\");");
        for (int i = 0; i < m.getColumnCount(); i++) {
            String column = m.getColumnName(i + 1);
            String property = "";
            JoinTable joinTable = joinTables.get(column);
            Class nestedPropertyType = null;
            
            try {
                try {
                    Query.propertyUtils.getProperty(prototype, column);
                    property = column;
                } catch (Exception ex) {
                    if (column.contains("_")) {
                        String fixedCase = replaceUnderscoreWithCase(column);
                        Query.propertyUtils.getProperty(prototype, fixedCase);
                        property = fixedCase;
                    } else {
                        throw ex;
                    }
                }
                if (joinTable != null) {
                    Query.ensureObjectsCreated(prototype, column);
                    nestedPropertyType = Query.propertyUtils.getPropertyType(prototype, column);
                    String nestedProperty = column + "." + joinTable.pkColumn;
                    Query.ensureObjectsCreated(prototype, column);
                    Query.propertyUtils.getProperty(prototype, nestedProperty);
                    property = property + "." + joinTable.pkColumn;
                }
            } catch (Exception noSuchField) {
                nestedPropertyType = null;
                property = "";
            }
            
            String fieldType = pks.contains(column) ? "id" : "add";

            s.append("\n\t" + fieldType + "(\"").append(property).append("\", \"").append(column).append("\"");
            if (m.isAutoIncrement(i + 1)) {
                ResultSet columnRs = dbm.getColumns(null, null, tablename, column);
                String seqName = "sequence_name";
                if (columnRs.next()) {
                    String defaultValue = columnRs.getString("COLUMN_DEF");
                    Pattern seq = Pattern.compile(".*nextval\\(\'(.*?)\'.*");
                    Matcher matcher = seq.matcher(defaultValue);
                    if (matcher.matches())
                        seqName = matcher.group(1);
                }

                s.append(", \"" + seqName + "\"");
            }
            s.append(");");
            if (nestedPropertyType != null)
                appendJoinTableSkeleton(dbm, nestedPropertyType, s, joinTable, connection);
        }
        s.append("\n}};\n");
        return s.toString();
    }

    private static String replaceUnderscoreWithCase(String s) {
        Pattern us = Pattern.compile("_([a-z]{1})");
        Matcher m = us.matcher(s.toLowerCase());
        StringBuffer sb = new StringBuffer();
        while (m.find()) {
            String c = m.group(1).toUpperCase();
            m.appendReplacement(sb, c);
        }
        m.appendTail(sb);
        return sb.toString();
    }

    private static void appendJoinTableSkeleton(DatabaseMetaData dbm, Class domainClass, StringBuilder s, JoinTable joinTable, Connection connection) throws SQLException, IllegalAccessException, InstantiationException {
        ResultSet rs = connection.createStatement().executeQuery("SELECT * FROM " + joinTable.getName());
        ResultSetMetaData m = rs.getMetaData();
        Object prototype = domainClass.newInstance();

        ResultSet importedKeys = dbm.getImportedKeys(null, null, joinTable.getName());
        List<String> joinColumns = new ArrayList<>();
        while (importedKeys.next())
            joinColumns.add(importedKeys.getString("FKCOLUMN_NAME"));

        for (int i = 0; i < m.getColumnCount(); i++) {
            String column = m.getColumnName(i + 1);
            String property = null;

            if (column.equals(joinTable.pkColumn) || joinColumns.contains(column))
                continue;

            try {
                Query.propertyUtils.getProperty(prototype, column);
                property = column;
            } catch (Exception noSuchField) {
            }

            String nestedProperty = joinTable.getAlias() + "." + (property == null ? "" : property);
            
            s.append("\n\tjoin(\"").append(nestedProperty).append("\", ").append(joinTable.getAlias()).append(", \"").append(column).append("\"");
            s.append(");");
        }        
    }

    public void merge(String columnPrefix, String propertyPrefix, ResultMap resultMap) {
        for (String key : resultMap.getMappings().keySet()) {
            ResultMapping mapping = resultMap.getMappings().get(key);
            mapping.setProperty(propertyPrefix + "." + mapping.getProperty());
            mappings.put(columnPrefix + "_" + key, mapping);
        }
    }

    public String columns(Boolean includeJoins) {
        StringBuilder s = new StringBuilder();
        for (ResultMapping mapping : mappings.values()) {
            if (includeJoins || (mapping.getMapType().equals(MapType.TABLE_COLUMN) || mapping.getMapType().equals(MapType.PRIMARY_KEY))) {
                if (s.length() > 0)
                    s.append(", ");

                if (mapping.getJoinTable() != null)
                    s.append(mapping.getJoinTable().getAlias() + "." + mapping.getColumn() + " AS " + mapping.getJoinTable().getAlias() + "_" + mapping.getColumn());
                else if (includeJoins)
                    s.append(tablename + "." + mapping.getColumn());
                else
                    s.append(mapping.getColumn());
            }
        }
        return s.toString();
    }

    public String values() {
        StringBuilder s = new StringBuilder();
        for (ResultMapping mapping : mappings.values()) {
            if (mapping.getMapType().equals(MapType.TABLE_COLUMN) || mapping.getMapType().equals(MapType.PRIMARY_KEY)) {
                if (s.length() > 0)
                    s.append(", ");
                s.append(":").append(mapping.getProperty());
            }
        }
        return s.toString();
    }

    public String select() {
        StringBuilder s = new StringBuilder("SELECT " + columns(true) + " FROM " + tablename);
        List<JoinTable> joinTables = uniqueJoinTables();

        // Direct joins
        for (JoinTable joinTable : joinTables) {
            if (joinTable.getJoinType().equals(JoinTable.JOIN_TYPE.JOIN)) {
                s.append(", ").append(joinTable.getName()).append(" AS ").append(joinTable.getAlias());
            }
        }

        // Left joins
        for (JoinTable joinTable : joinTables) {
            if (joinTable.getJoinType().equals(JoinTable.JOIN_TYPE.LEFT_JOIN))
                s.append(" LEFT JOIN ").append(joinTable.getName()).append(" AS ").append(joinTable.getAlias())
                        .append(" ON " + joinTable.getOn());
        }

        // Where clauses for direct joins
        boolean first = true;
        for (JoinTable joinTable : joinTables) {
            if (joinTable.getJoinType().equals(JoinTable.JOIN_TYPE.JOIN)) {
                if (first) {
                    first = false;
                    s.append(" WHERE ");
                } else {
                    s.append(" AND ");
                }
                s.append(joinTable.getOn());
            }
        }

        return s.toString();
    }

    public List<JoinTable> uniqueJoinTables() {
        List<JoinTable> joinTables = new ArrayList<>();
        for (ResultMapping mapping : mappings.values()) {
            if (mapping.getJoinTable() != null && !joinTables.contains(mapping.getJoinTable()))
                joinTables.add(mapping.getJoinTable());
        }
        return joinTables;
    }

    public String insert() {
        return "INSERT INTO " + tablename + " (" + columns(false) + ") VALUES (" + values() + ")";
    }

    public String updateStatement(String param) {
        StringBuilder s = new StringBuilder();
        for (ResultMapping mapping : mappings.values()) {
            if (mapping.getMapType().equals(MapType.TABLE_COLUMN)) {
                if (s.length() > 0)
                    s.append(", ");
                s.append(mapping.getColumn()).append(" = ").append(":");
                if (param != null)
                    s.append(param + ".");

                s.append(mapping.getProperty());
            }
        }
        return s.toString();
    }

    public String update() {
        return "UPDATE " + tablename + " SET " + updateStatement(null) + " WHERE " + identifyColumn();
    }

    private String identifyColumn() {
        StringBuilder s = new StringBuilder();
        for (ResultMapping mapping : mappings.values()) {
            if (mapping.getMapType().equals(MapType.PRIMARY_KEY)) {
                if (s.length() > 0)
                    s.append("AND ");
                s.append(mapping.getColumn()).append("=:").append(mapping.getProperty());
            }
        }
        return s.toString();
    }

    public String update(String param) {
        return "UPDATE " + tablename + " SET " + updateStatement(param) + " WHERE " + identifyColumn();
    }

    public void setMappings(Map<String, ResultMapping> mappings) {
        this.mappings = mappings;
    }

    public boolean containsColumn(String columnName) {
        return mappings.containsKey(columnName);
    }

    public ResultMap(Class resultClass) {
        this.resultClass = resultClass;
    }

    public ResultMap(Class resultClass, Boolean strict) {
        this.resultClass = resultClass;
        this.strict = strict;
    }

    public ResultMap(Class resultClass, Boolean strict, Boolean automap) {
        this.resultClass = resultClass;
        this.strict = strict;
        this.automap = automap;
    }

    public ResultMap add(String property, String column) {
        mappings.put(column, new ResultMapping(property, column));
        return this;
    }

    public ResultMap join(String property, String column) {
        return add(property, column, MapType.JOIN_COLUMN);
    }

    public ResultMap join(String property, JoinTable joinTable, String column) {
        return add(new ResultMapping(property, column, MapType.JOIN_COLUMN).joinTable(joinTable), joinTable.getAlias() + "_" + column);
    }

    public ResultMap add(String property, String column, MapType mapType) {
        mappings.put(column, new ResultMapping(property, column, mapType));
        return this;
    }

    public ResultMap id(String property, String column) {
        mappings.put(column, new ResultMapping(property, column, MapType.PRIMARY_KEY));
        return this;
    }

    public ResultMap id(String property, String column, String sequence) {
        ResultMapping mapping = new ResultMapping(property, column, MapType.PRIMARY_KEY);
        mapping.setSequence(sequence);
        mappings.put(column, mapping);
        return this;
    }

    public ResultMap add(ResultMapping mapping, String column) {
        mappings.put(column, mapping);
        return this;
    }

    public Class getResultClass() {
        return resultClass;
    }

    public void setResultClass(Class resultClass) {
        this.resultClass = resultClass;
    }

    public Boolean getStrict() {
        return strict;
    }

    public void setStrict(Boolean strict) {
        this.strict = strict;
    }

    public ResultMap strict() {
        this.strict = true;
        return this;
    }

    public ResultMap lenient() {
        this.strict = false;
        return this;
    }

    public String getTablename() {
        return tablename;
    }

    public ResultMap tablename(String tablename) {
        this.tablename = tablename;
        return this;
    }

    public Boolean getAutomap() {
        return automap;
    }

    public void setAutomap(Boolean automap) {
        this.automap = automap;
    }
    
    ResultMap automap(boolean automap) {
        setAutomap(automap);
        return this;
    }
}
